"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.deserializePayload = exports.deserialize = exports.payloadDiscriminator = exports.serializePayload = exports.serialize = exports.registerPayloadType = exports.create = void 0;
const sdk_base_1 = require("@wormhole-foundation/sdk-base");
const layout_items_1 = require("./layout-items");
const utils_1 = require("./utils");
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...layout_items_1.signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...layout_items_1.guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        layout: guardianSignatureLayout,
    },
];
//envelope + payload are getting hashed and signed
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...(0, layout_items_1.chainItem)() },
    { name: "emitterAddress", ...layout_items_1.universalAddressItem },
    { name: "sequence", ...layout_items_1.sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
const payloadFactory = new Map();
function getPayloadLayout(layoutLiteral) {
    const layout = payloadFactory.get(layoutLiteral);
    if (!layout)
        throw new Error(`No layout registered for payload type ${layoutLiteral}`);
    return layout;
}
const payloadLiteralToPayloadItem = (payloadLiteral) => (payloadLiteral === "Uint8Array"
    ? { name: "payload", binary: "bytes" }
    : {
        name: "payload",
        binary: "object",
        layout: getPayloadLayout(payloadLiteral),
    });
const create = (payloadLiteral, vaaData) => {
    const bodyLayout = [
        ...envelopeLayout,
        payloadLiteralToPayloadItem(payloadLiteral),
    ];
    const bodyWithFixed = (0, sdk_base_1.addFixedValues)(bodyLayout, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        payloadLiteral,
        ...(0, sdk_base_1.addFixedValues)(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: (0, utils_1.keccak256)((0, sdk_base_1.serializeLayout)(bodyLayout, bodyWithFixed)),
    };
};
exports.create = create;
function registerPayloadType(payloadLiteral, payloadLayout) {
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, payloadLayout);
}
exports.registerPayloadType = registerPayloadType;
const serialize = (vaa) => {
    const layout = [
        ...baseLayout,
        payloadLiteralToPayloadItem(vaa.payloadLiteral),
    ];
    return (0, sdk_base_1.serializeLayout)(layout, vaa);
};
exports.serialize = serialize;
const serializePayload = (payloadLiteral, payload) => {
    if (payloadLiteral === "Uint8Array")
        return payload;
    const layout = getPayloadLayout(payloadLiteral);
    return (0, sdk_base_1.serializeLayout)(layout, payload);
};
exports.serializePayload = serializePayload;
const payloadDiscriminator = (namedPayloads) => {
    const literals = (0, sdk_base_1.column)(namedPayloads, 0);
    const layouts = (0, sdk_base_1.column)(namedPayloads, 1);
    const discriminator = (0, sdk_base_1.layoutDiscriminator)(layouts);
    return (data) => {
        if (typeof data === "string")
            data = (0, sdk_base_1.hexByteStringToUint8Array)(data);
        const index = discriminator(data);
        return index !== null ? literals[index] : null;
    };
};
exports.payloadDiscriminator = payloadDiscriminator;
function deserialize(payloadDet, data) {
    if (typeof data === "string")
        data = (0, sdk_base_1.hexByteStringToUint8Array)(data);
    const [header, envelopeOffset] = (0, sdk_base_1.deserializeLayout)(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const [envelope, payloadOffset] = (0, sdk_base_1.deserializeLayout)(envelopeLayout, data, envelopeOffset, false);
    const [payloadLiteral, payload] = payloadDet === "Uint8Array"
        ? ["Uint8Array", data.slice(payloadOffset)]
        : typeof payloadDet === "string"
            ? [
                payloadDet,
                (0, sdk_base_1.deserializeLayout)(getPayloadLayout(payloadDet), data, payloadOffset),
            ]
            : deserializePayload(payloadDet, data, payloadOffset);
    const hash = (0, utils_1.keccak256)(data.slice(envelopeOffset));
    return { payloadLiteral, ...header, ...envelope, payload, hash };
}
exports.deserialize = deserialize;
function deserializePayload(payloadDet, data, offset = 0) {
    //grouped together to have just a single cast on the return type
    return (() => {
        if (typeof data === "string")
            data = (0, sdk_base_1.hexByteStringToUint8Array)(data);
        if (payloadDet === "Uint8Array")
            return data.slice(offset);
        if (typeof payloadDet === "string")
            return (0, sdk_base_1.deserializeLayout)(getPayloadLayout(payloadDet), data, offset);
        const candidate = payloadDet(data);
        if (candidate === null)
            throw new Error(`Encoded data does not match any of the given payload types - ${data}`);
        return [
            candidate,
            (0, sdk_base_1.deserializeLayout)(getPayloadLayout(candidate), data, offset),
        ];
    })();
}
exports.deserializePayload = deserializePayload;
//# sourceMappingURL=vaa.js.map