import { column, hexByteStringToUint8Array, layoutDiscriminator, serializeLayout, deserializeLayout, addFixedValues, } from "@wormhole-foundation/sdk-base";
import { chainItem, universalAddressItem, signatureItem, sequenceItem, guardianSetItem, } from "./layout-items";
import { keccak256 } from "./utils";
const guardianSignatureLayout = [
    { name: "guardianIndex", binary: "uint", size: 1 },
    { name: "signature", ...signatureItem },
];
const headerLayout = [
    { name: "version", binary: "uint", size: 1, custom: 1, omit: true },
    { name: "guardianSet", ...guardianSetItem },
    {
        name: "signatures",
        binary: "array",
        lengthSize: 1,
        layout: guardianSignatureLayout,
    },
];
//envelope + payload are getting hashed and signed
const envelopeLayout = [
    { name: "timestamp", binary: "uint", size: 4 },
    { name: "nonce", binary: "uint", size: 4 },
    { name: "emitterChain", ...chainItem() },
    { name: "emitterAddress", ...universalAddressItem },
    { name: "sequence", ...sequenceItem },
    { name: "consistencyLevel", binary: "uint", size: 1 },
];
const baseLayout = [...headerLayout, ...envelopeLayout];
const payloadFactory = new Map();
function getPayloadLayout(layoutLiteral) {
    const layout = payloadFactory.get(layoutLiteral);
    if (!layout)
        throw new Error(`No layout registered for payload type ${layoutLiteral}`);
    return layout;
}
const payloadLiteralToPayloadItem = (payloadLiteral) => (payloadLiteral === "Uint8Array"
    ? { name: "payload", binary: "bytes" }
    : {
        name: "payload",
        binary: "object",
        layout: getPayloadLayout(payloadLiteral),
    });
export const create = (payloadLiteral, vaaData) => {
    const bodyLayout = [
        ...envelopeLayout,
        payloadLiteralToPayloadItem(payloadLiteral),
    ];
    const bodyWithFixed = addFixedValues(bodyLayout, 
    //not sure why the unknown cast here is required and why the type isn't inferred correctly
    vaaData);
    return {
        payloadLiteral,
        ...addFixedValues(headerLayout, vaaData),
        ...bodyWithFixed,
        hash: keccak256(serializeLayout(bodyLayout, bodyWithFixed)),
    };
};
export function registerPayloadType(payloadLiteral, payloadLayout) {
    if (payloadFactory.has(payloadLiteral))
        throw new Error(`Payload type ${payloadLiteral} already registered`);
    payloadFactory.set(payloadLiteral, payloadLayout);
}
export const serialize = (vaa) => {
    const layout = [
        ...baseLayout,
        payloadLiteralToPayloadItem(vaa.payloadLiteral),
    ];
    return serializeLayout(layout, vaa);
};
export const serializePayload = (payloadLiteral, payload) => {
    if (payloadLiteral === "Uint8Array")
        return payload;
    const layout = getPayloadLayout(payloadLiteral);
    return serializeLayout(layout, payload);
};
export const payloadDiscriminator = (namedPayloads) => {
    const literals = column(namedPayloads, 0);
    const layouts = column(namedPayloads, 1);
    const discriminator = layoutDiscriminator(layouts);
    return (data) => {
        if (typeof data === "string")
            data = hexByteStringToUint8Array(data);
        const index = discriminator(data);
        return index !== null ? literals[index] : null;
    };
};
export function deserialize(payloadDet, data) {
    if (typeof data === "string")
        data = hexByteStringToUint8Array(data);
    const [header, envelopeOffset] = deserializeLayout(headerLayout, data, 0, false);
    //ensure that guardian signature indicies are unique and in ascending order - see:
    //https://github.com/wormhole-foundation/wormhole/blob/8e0cf4c31f39b5ba06b0f6cdb6e690d3adf3d6a3/ethereum/contracts/Messages.sol#L121
    for (let i = 1; i < header.signatures.length; ++i)
        if (header.signatures[i].guardianIndex <=
            header.signatures[i - 1].guardianIndex)
            throw new Error("Guardian signatures must be in ascending order of guardian set index");
    const [envelope, payloadOffset] = deserializeLayout(envelopeLayout, data, envelopeOffset, false);
    const [payloadLiteral, payload] = payloadDet === "Uint8Array"
        ? ["Uint8Array", data.slice(payloadOffset)]
        : typeof payloadDet === "string"
            ? [
                payloadDet,
                deserializeLayout(getPayloadLayout(payloadDet), data, payloadOffset),
            ]
            : deserializePayload(payloadDet, data, payloadOffset);
    const hash = keccak256(data.slice(envelopeOffset));
    return { payloadLiteral, ...header, ...envelope, payload, hash };
}
export function deserializePayload(payloadDet, data, offset = 0) {
    //grouped together to have just a single cast on the return type
    return (() => {
        if (typeof data === "string")
            data = hexByteStringToUint8Array(data);
        if (payloadDet === "Uint8Array")
            return data.slice(offset);
        if (typeof payloadDet === "string")
            return deserializeLayout(getPayloadLayout(payloadDet), data, offset);
        const candidate = payloadDet(data);
        if (candidate === null)
            throw new Error(`Encoded data does not match any of the given payload types - ${data}`);
        return [
            candidate,
            deserializeLayout(getPayloadLayout(candidate), data, offset),
        ];
    })();
}
//# sourceMappingURL=vaa.js.map